/***************************************************************************
Copyright (c) 2016, Guillaume Baud-Berthier, SafeRiver/LaBRI

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
IN THE SOFTWARE.
***************************************************************************/

#include "bmc.h"

unsigned act_init;

std::vector<int> shift_spod;
std::vector<int> var2cycle;


dll_list<ands_def>* unroll_kind(int time) {
  int j;
  unsigned tmp_rhs;
  unsigned tmp_rhs0, tmp_rhs1;

  dll_list<ands_def> *tmp_list_ands;
  tmp_list_ands = new dll_list<ands_def>();
  dll_init<ands_def>(tmp_list_ands);
  
  idef* tmp_def;
  dll_elem<idef> *tmp_e;
  std::list<unsigned>::iterator it_rhs;
  std::set<unsigned>::iterator it_tmp_rhs;


  // Inputs
  tmp_e = dll_inputs->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    tmp_def->lhs_ts.push_back(new_var());
    var2cycle.push_back(time);
    tmp_e = tmp_e->next;
  }  
  
  // Latches
  tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    if (!time) { // Init
      tmp_def->lhs_ts.push_back(new_var());
      var2cycle.push_back(time);
      bcl(-(tmp_def->lhs_ts[0] >> 1), -act_init);

    } else {
      assert(tmp_def->rhs.size() == 1);
      
      if (tmp_def->rhs.front() > 1) {
	assert(tmp_def->lhs_ts[time-1] != -1);

	assert(tmp_def->lhs_ts.size() == time);
	tmp_def->lhs_ts.push_back(new_var());
	var2cycle.push_back(time);

	if (aiger_sign(tmp_def->rhs.front()) != aiger_sign(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1])) {
	  bcl( (tmp_def->lhs_ts[time] >> 1),  (mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	  bcl(-(tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	} else {
	  bcl( (tmp_def->lhs_ts[time] >> 1), -(mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1));
	  bcl(-(tmp_def->lhs_ts[time] >> 1),   mapdef[aiger_lit2var(tmp_def->rhs.front())]->lhs_ts[time-1] >> 1);
	}
	

      } else { // Cst latch
	tmp_def->lhs_ts.push_back(new_var());
	var2cycle.push_back(time);
	assert(tmp_def->rhs.front() == 1);
	ucl(tmp_def->lhs_ts[time] >> 1);	
      }
    }
    tmp_e = tmp_e->next;
  }
  
  // Ands
  tmp_e = dll_ands->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    prop_tr_un *tpu = unroll_prop_tr(tmp_def->rhs_tr, time, 0);

    if (tpu->lit == UNDEF_LIT) {
      assert(tpu->op != VAR);

      tmp_def->lhs_ts.push_back(new_var());
      var2cycle.push_back(time);

      ands_def *tmp_ands = new ands_def(tmp_def->lhs_ts[time], tpu, tmp_def, time);

      map_ands.insert(std::pair<unsigned, ands_def*>(aiger_lit2var(tmp_ands->lhs), tmp_ands));
      
      tmp_ands->dllade = dll_insert_last(tmp_list_ands, aiger_lit2var(tmp_def->lhs_ts[time]), tmp_ands);

    } else {
      assert(tpu->op == VAR);

      tmp_def->lhs_ts.push_back(new_var());
      var2cycle.push_back(time);

      if (tpu->lit == 0) {
	ucl(-(tmp_def->lhs_ts[time] >> 1));

      } else if (tpu->lit == 1) {
	ucl(tmp_def->lhs_ts[time] >> 1);

      } else if (aiger_sign(tpu->lit)) {
	bcl( (tmp_def->lhs_ts[time] >> 1),  (tpu->lit >> 1));
	bcl(-(tmp_def->lhs_ts[time] >> 1), -(tpu->lit >> 1));

      } else {
	bcl( (tmp_def->lhs_ts[time] >> 1), -(tpu->lit >> 1));
	bcl(-(tmp_def->lhs_ts[time] >> 1),  (tpu->lit >> 1));
      }
      
    }

    tmp_e = tmp_e->next;
  }

  return tmp_list_ands;
}



void replicate_clauses(int current_cycle, bool rep_in_learntDB, bool verbose_) {
  vec<Lit> lits;
  cl_rep cl;
  int diff_cycle;
  int stats_rep = 0;
  for(int i = 0; i < solver->cl_rep_vec.size(); i++) {
    cl = solver->cl_rep_vec[i];
    diff_cycle = current_cycle - cl.learnt_cycle;
    bool rep = 1;
    for (int j = 0; j < cl.size_cl && rep; j++) {
      int cycle_var = var2cycle[var(cl.lits[j])];
      assert(cycle_var != -1);
      if (cycle_var == -2)
      	rep = 0;

      if (rep) {
	int var_wo_shift = var(cl.lits[j]) - shift_spod[cycle_var];
	lits.push(mkLit(var_wo_shift + (solver->nb_vars_by_cycle * diff_cycle) +
			shift_spod[cycle_var + diff_cycle], sign(cl.lits[j])));
      }
    }
    if (rep) {
      if (rep_in_learntDB)
	solver->addClause_learntDB(lits, cl.lbd);
      else
	solver->Solver::addClause_(lits);
      stats_rep++;
    }
    lits.clear();
  }

  if (verbose_)
    if (solver->rep_max_sz < INT_MAX)
      printf("[zigzag]  %d/%d clauses (max size %d) replicated to new cycle!\n", stats_rep,
	     solver->cl_rep_vec.size(), solver->rep_max_sz);
    else
      printf("[zigzag]  %d/%d clauses replicated to new cycle!\n", stats_rep, solver->cl_rep_vec.size());
}


void replicate_activity_kind_v1(int i) {
  idef* tmp_def;
  dll_elem<idef> *tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    solver->activity[tmp_def->lhs_ts[0] >> 1] = 1e90;
      
    if (solver->order_heap.inHeap(tmp_def->lhs_ts[0] >> 1))
      solver->order_heap.decrease(tmp_def->lhs_ts[0] >> 1);
      
    tmp_e = tmp_e->next;
  }
}

void replicate_activity_kind_v2(int i) {
  idef* tmp_def;
  dll_elem<idef> *tmp_e = dll_latches->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    solver->activity[tmp_def->lhs_ts[0] >> 1] = 1e90;
      
    if (solver->order_heap.inHeap(tmp_def->lhs_ts[0] >> 1))
      solver->order_heap.decrease(tmp_def->lhs_ts[0] >> 1);
      
    tmp_e = tmp_e->next;
  }
    
  tmp_e = dll_ands->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;
    
    for (int j = 0; j < i; j++) {
      
      solver->activity[tmp_def->lhs_ts[j] >> 1] =
	(solver->varBumpCounter[tmp_def->lhs_ts[j] >> 1] + solver->old_varBumpCounter[tmp_def->lhs_ts[j] >> 1]) / 2;

      if (solver->order_heap.inHeap(tmp_def->lhs_ts[j] >> 1))
	solver->order_heap.decrease(tmp_def->lhs_ts[j] >> 1);
    }
    
    solver->activity[tmp_def->lhs_ts[i] >> 1] = solver->varBumpCounter[tmp_def->lhs_ts[i-1] >> 1] / 2;

    if (solver->order_heap.inHeap(tmp_def->lhs_ts[i] >> 1))
      solver->order_heap.decrease(tmp_def->lhs_ts[i] >> 1);

    tmp_e = tmp_e->next;
  }

  
  tmp_e = dll_inputs->head;
  while(tmp_e != NULL) {
    tmp_def = tmp_e->def;

    
    for (int j = 0; j < i; j++) {
      
      solver->activity[tmp_def->lhs_ts[j] >> 1] =
	(solver->varBumpCounter[tmp_def->lhs_ts[j] >> 1] + solver->old_varBumpCounter[tmp_def->lhs_ts[j] >> 1]) / 2;

      if (solver->order_heap.inHeap(tmp_def->lhs_ts[j] >> 1))
	solver->order_heap.decrease(tmp_def->lhs_ts[j] >> 1);
    }
    
    solver->activity[tmp_def->lhs_ts[i] >> 1] = solver->varBumpCounter[tmp_def->lhs_ts[i-1] >> 1] / 2;

    if (solver->order_heap.inHeap(tmp_def->lhs_ts[i] >> 1))
      solver->order_heap.decrease(tmp_def->lhs_ts[i] >> 1);


    tmp_e = tmp_e->next;
  }
}



int bmc(int start, int stop, aiger *model, bool elim_hl, int simple, int rep, bool repG, int rep_max_sz,
	bool verbose_) {
  int i, j;

  int status = -1;

  unsigned bad_lit = init_model(model, stop, 0, verbose_);

  if (elim_hl)
    hl_ve(verbose_);

  var2cycle.push_back(-3);

  act_init = new_var() >> 1;
  var2cycle.push_back(-1);
  int nb_vars_by_cycle = dll_latches->sz + dll_inputs->sz + dll_ands->sz;

  solver->nb_vars_by_cycle = nb_vars_by_cycle;
  solver->rep_max_sz = rep_max_sz;
  solver->repGlue = repG;
  solver->bad_lit = act_init;

  shift_spod.push_back(0);

  if (!rep)
    ucl(act_init);

  for(i = 0; i <= stop; i++) {
    solver->current_cycle = i;

    shift_spod.push_back(shift_spod[i]);

    if (verbose_) printf("[zigzag]  Encoding %d..\n", i);

    list_ands.push_back(unroll_kind(i));

    iter_flatten_ptu(i);

    ands2clauses(i);

    if (rep == 1) // Replicate in clause DB, i.e. cannot be removed
      replicate_clauses(i, 0, verbose_);
    else if (rep == 2) // Replicate in learnt clauses DB
      replicate_clauses(i, 1, verbose_);

    assert(mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i] > 1);
    
    int bad = mapdef[aiger_lit2var(bad_lit)]->lhs_ts[i];
    bad = aiger_sign(bad_lit) == aiger_sign(bad) ? (bad >> 1) : -(bad >> 1);

    lbool res_check;
    if (verbose_) printf("[zigzag]  Solving BMC %d..\n", i);

    assume(bad);              // Assume: -P(x)
    if (rep) {
      assume(act_init);       // Assume:  I(x)
    }

    res_check = check_sat(verbose_);
    if (res_check == l_True) {
      status = 1;
      break;
    }


    printf("u%d\n", i);
    fflush(stdout);

    ucl(-bad);                // Add: P(x)

  }

  printf("%d\nb0\n.\n", status);
  fflush(stdout);

  if (verbose_)
    printf("Vars %d (%d - %d)| Clauses %d\n", solver->nVars() - solver->eliminated_vars,
	   solver->nVars(), solver->eliminated_vars, solver->nClauses());

  if (verbose_)
    printf("Learnt : %d | STATS REPLICABLE FOUND : %d\n",
	   solver->nLearnts(), solver->stats_replicable_clause_found);

  destroy_model(verbose_);
  return 0;
}

